/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

/*  
 Reasoning behind the mathematics in this module (in the key of 'Mathematica'):
 
 s == speed, a == acceleration, t == time, d == distance
 
 Basic definitions:
 
 Speed[s_, a_, t_] := s + (a*t) 
 Travel[s_, a_, t_] := Integrate[Speed[s, a, t], t]
 
 Distance to reach a specific speed with a constant acceleration:
 
 Solve[{Speed[s, a, t] == m, Travel[s, a, t] == d}, d, t]
 d -> (m^2 - s^2)/(2 a) --> estimate_acceleration_distance()
 
 Speed after a given distance of travel with constant acceleration:
 
 Solve[{Speed[s, a, t] == m, Travel[s, a, t] == d}, m, t]
 m -> Sqrt[2 a d + s^2]    
 
 DestinationSpeed[s_, a_, d_] := Sqrt[2 a d + s^2]
 
 When to start braking (di) to reach a specified destionation speed (s2) after accelerating
 from initial speed s1 without ever stopping at a plateau:
 
 Solve[{DestinationSpeed[s1, a, di] == DestinationSpeed[s2, a, d - di]}, di]
 di -> (2 a d - s1^2 + s2^2)/(4 a) --> intersection_distance()
 
 IntersectionDistance[s1_, s2_, a_, d_] := (2 a d - s1^2 + s2^2)/(4 a)
 */

#include "MarlinHOH.h"
#include "planner.h"
#include "stepper.h"
#include "temperature.h"
#include "language.h"
#include "parameter.h"
#include "printer.h"
#include "fastio.h"


#ifdef PREVENT_DANGEROUS_EXTRUDE
float extrude_min_temp=EXTRUDE_MINTEMP;
#endif
#ifdef XY_FREQUENCY_LIMIT
#define MAX_FREQ_TIME (1000000.0/XY_FREQUENCY_LIMIT)
// Used for the frequency limit
static unsigned char old_direction_bits = 0;               // Old direction bits. Used for speed calculations
static long x_segment_time[3]={MAX_FREQ_TIME + 1,0,0};     // Segment times (in us). Used for speed calculations
static long y_segment_time[3]={MAX_FREQ_TIME + 1,0,0};
#endif
#ifdef PRESSURE_ADVANCE
float last_e = 0;
#endif

float junction_deviation = 0.1;

#ifdef PREVENT_DANGEROUS_EXTRUDE
void set_extrude_min_temp(float temp)
{
  extrude_min_temp=temp;
}
#endif

Planner::Planner()
{
}

Planner::~Planner()
{
}

void Planner::Init()
{
	block_buffer_head = 0;
	block_buffer_tail = 0;
	memset(position_st, 0, sizeof(position_st)); // clear position_st
	previous_speed[0] = 0.0;
	previous_speed[1] = 0.0;
	previous_speed[2] = 0.0;
	previous_speed[3] = 0.0;
	previous_nominal_speed = 0.0;
}

void Planner::Set_Position_mm(const float& x, const float& y, const float& z,
		const float& e)
{
  position_st[X_AXIS] = lround(x*config.axis_steps_per_unit[X_AXIS]);
  position_st[Y_AXIS] = lround(y*config.axis_steps_per_unit[Y_AXIS]);
  position_st[Z_AXIS] = lround(z*config.axis_steps_per_unit[Z_AXIS]);
  position_st[E_AXIS] = lround(e*config.axis_steps_per_unit[E_AXIS]);
  motor.Set_Position(position_st[X_AXIS], position_st[Y_AXIS], position_st[Z_AXIS], position_st[E_AXIS]);
  previous_nominal_speed = 0.0; // Resets planner junction speeds. Assumes start from rest.
  previous_speed[0] = 0.0;
  previous_speed[1] = 0.0;
  previous_speed[2] = 0.0;
  previous_speed[3] = 0.0;
#ifdef PRESSURE_ADVANCE
  last_e = e;
#endif

}

void Planner::Check_Axes_Activity()
{
	unsigned char x_active = 0;
	unsigned char y_active = 0;
	unsigned char z_active = 0;
	unsigned char e_active = 0;
	unsigned char tail_fan_speed = heater.fanSpeed;
#ifdef BARICUDA
	unsigned char tail_valve_pressure = ValvePressure;
	unsigned char tail_e_to_p_pressure = EtoPPressure;
#endif
	move_block *block;

	if(block_buffer_tail != block_buffer_head)
	{
		uint8_t block_index = block_buffer_tail;
		tail_fan_speed = block_buffer[block_index].fan_speed;
#ifdef BARICUDA
		tail_valve_pressure = block_buffer[block_index].valve_pressure;
		tail_e_to_p_pressure = block_buffer[block_index].e_to_p_pressure;
#endif
		while(block_index != block_buffer_head)
		{
			block = &block_buffer[block_index];
			if(block->steps[X_AXIS]  != 0) x_active++;
			if(block->steps[Y_AXIS]  != 0) y_active++;
			if(block->steps[Z_AXIS]  != 0) z_active++;
			if(block->steps[E_AXIS]  != 0) e_active++;
			block_index = (block_index+1) & (BLOCK_BUFFER_SIZE - 1);
		}
	}
	if((DISABLE_X) && (x_active == 0)) disable_X();
	if((DISABLE_Y) && (y_active == 0)) disable_Y();
	if((DISABLE_Z) && (z_active == 0)) disable_Z();
	if((DISABLE_E) && (e_active == 0))
	{
		disable_E0();
		disable_E1();
		disable_E2();
	}
#if defined(FAN_PIN) && FAN_PIN > -1
	Write_Analogue(FAN_PIN,tail_fan_speed);
#endif//FAN_PIN > -1

#ifdef BARICUDA
#if defined(HEATER_1_PIN) && HEATER_1_PIN > -1
	analogWrite(HEATER_1_PIN,tail_valve_pressure);
#endif

#if defined(HEATER_2_PIN) && HEATER_2_PIN > -1
	analogWrite(HEATER_2_PIN,tail_e_to_p_pressure);
#endif
#endif
}

void Planner::Set_Position_mm(const AxisEnum axis, const float& value)
{
	  position_st[axis] = lround(value*config.axis_steps_per_unit[axis]);
	  motor.Set_Position(axis, position_st[axis]);
	#ifdef PRESSURE_ADVANCE
	  if (axis == E_AXIS)
	  last_e = value;
	#endif
}

void Planner::Buffer_Line(const float& x, const float& y, const float& z,
		const float& e, float feed_rate, const uint8_t& extruder)
{
	  // Calculate the buffer head after we push this byte
	  int next_buffer_head = Next_Block_Index(block_buffer_head);

	  // If the buffer is full: good! That means we are well ahead of the robot.
	  // Rest here until there is room in the buffer.
	  while(block_buffer_tail == next_buffer_head)
	  {
      	my_printer.Go_Idle();
	  }

	#ifdef ENABLE_AUTO_BED_LEVELING
	  apply_rotation_xyz(plan_bed_level_matrix, x, y, z);
	#endif // ENABLE_AUTO_BED_LEVELING

	  // The target position of the tool in absolute steps
	  // Calculate target position in absolute steps
	  //this should be done after the wait, because otherwise a M92 code within the gcode disrupts this calculation somehow
	  long target[4];
	  target[X_AXIS] = lround(x*config.axis_steps_per_unit[X_AXIS]);
	  target[Y_AXIS] = lround(y*config.axis_steps_per_unit[Y_AXIS]);
	  target[Z_AXIS] = lround(z*config.axis_steps_per_unit[Z_AXIS]);
	  target[E_AXIS] = lround(e*config.axis_steps_per_unit[E_AXIS]);

	  #ifdef PREVENT_DANGEROUS_EXTRUDE
	  if(target[E_AXIS]!=position_st[E_AXIS])
	  {
	    if(heater.degHotend(active_extruder)<extrude_min_temp)
	    {
	      position_st[E_AXIS]=target[E_AXIS]; //behave as if the move really took place, but ignore E part
	      my_printer.echo_ln(MSG_ERR_COLD_EXTRUDE_STOP);
	    }

	    #ifdef PREVENT_LENGTHY_EXTRUDE
	    if(labs(target[E_AXIS]-position_st[E_AXIS])>config.axis_steps_per_unit[E_AXIS]*EXTRUDE_MAXLENGTH)
	    {
	      position_st[E_AXIS]=target[E_AXIS]; //behave as if the move really took place, but ignore E part
	      my_printer.echo_ln(MSG_ERR_LONG_EXTRUDE_STOP);
	    }
	    #endif
	  }
	  #endif

	  // Prepare to set up new block
	  move_block *block = &block_buffer[block_buffer_head];
	  // Mark block as not busy (Not executed by the stepper interrupt)
	  block->busy = false;

	  // Number of steps for each axis
	#ifndef COREXY
	// default non-h-bot planning
	block->steps[X_AXIS]  = labs(target[X_AXIS]-position_st[X_AXIS]);
	block->steps[Y_AXIS]  = labs(target[Y_AXIS]-position_st[Y_AXIS]);
	#else
	// corexy planning
	// these equations follow the form of the dA and dB equations on http://www.corexy.com/theory.html
	block->steps_x = labs((target[X_AXIS]-position[X_AXIS]) + (target[Y_AXIS]-position[Y_AXIS]));
	block->steps_y = labs((target[X_AXIS]-position[X_AXIS]) - (target[Y_AXIS]-position[Y_AXIS]));
	#endif
	  block->steps[Z_AXIS]  = labs(target[Z_AXIS]-position_st[Z_AXIS]);
	  block->steps[E_AXIS]  = labs(target[E_AXIS]-position_st[E_AXIS]);
	  block->steps[E_AXIS] *= volumetric_multiplier[active_extruder];
	  block->steps[E_AXIS] *= extrudemultiply;
	  block->steps[E_AXIS] /= 100;
	  block->step_event_count = max(block->steps[X_AXIS], max(block->steps[Y_AXIS], max(block->steps[Z_AXIS], block->steps[E_AXIS])));

	  // Bail if this is a zero-length block
	  if (block->step_event_count <= dropsegments)
	  {
	    return;
	  }

	  block->fan_speed = heater.fanSpeed;
	  #ifdef BARICUDA
	  block->valve_pressure = ValvePressure;
	  block->e_to_p_pressure = EtoPPressure;
	  #endif

	  // Compute direction bits for this block
	  block->direction_bits = 0;
	#ifndef COREXY
	  if (target[X_AXIS] < position_st[X_AXIS])
	  {
	    block->direction_bits |= (1<<X_AXIS);
	  }
	  if (target[Y_AXIS] < position_st[Y_AXIS])
	  {
	    block->direction_bits |= (1<<Y_AXIS);
	  }
	#else
	  if ((target[X_AXIS]-position[X_AXIS]) + (target[Y_AXIS]-position[Y_AXIS]) < 0)
	  {
	    block->direction_bits |= (1<<X_AXIS);
	  }
	  if ((target[X_AXIS]-position[X_AXIS]) - (target[Y_AXIS]-position[Y_AXIS]) < 0)
	  {
	    block->direction_bits |= (1<<Y_AXIS);
	  }
	#endif
	  if (target[Z_AXIS] < position_st[Z_AXIS])
	  {
	    block->direction_bits |= (1<<Z_AXIS);
	  }
	  if (target[E_AXIS] < position_st[E_AXIS])
	  {
	    block->direction_bits |= (1<<E_AXIS);
	  }

	  block->active_extruder = extruder;

	#ifdef PRESSURE_ADVANCE
	// set the pressure advance factor for the acvtive extruder
	//block->advance = advance_factor[extruder];
	block->initial_advance = 0;
	block->final_advance = 0;
	block->advance_delay = config.advance_delay[extruder];
	#endif
	  //enable active axes
	  #ifdef COREXY
	  if((block->steps_x != 0) || (block->steps_y != 0))
	  {
	    enable_x();
	    enable_y();
	  }
	  #else
	  if(block->steps[X_AXIS] != 0) enable_X();
	  if(block->steps[Y_AXIS] != 0) enable_Y();
	  #endif
	#ifndef Z_LATE_ENABLE
	  if(block->steps[Z_AXIS] != 0) enable_Z();
	#endif

	  // Enable all
	  if(block->steps[E_AXIS] != 0)
	  {
	    enable_E0();
	    enable_E1();
	    enable_E2();
	  }

	  if (block->steps[E_AXIS] == 0)
	  {
	    if(feed_rate<config.min_travel_feedrate_mm_s) feed_rate=config.min_travel_feedrate_mm_s;
	  }
	  else
	  {
	    if(feed_rate<config.min_feedrate_mm_s) feed_rate=config.min_feedrate_mm_s;
	  }

	  float delta_mm[4];
	  #ifndef COREXY
	    delta_mm[X_AXIS] = (target[X_AXIS]-position_st[X_AXIS])/config.axis_steps_per_unit[X_AXIS];
	    delta_mm[Y_AXIS] = (target[Y_AXIS]-position_st[Y_AXIS])/config.axis_steps_per_unit[Y_AXIS];
	  #else
	    delta_mm[X_AXIS] = ((target[X_AXIS]-position[X_AXIS]) + (target[Y_AXIS]-position[Y_AXIS]))/axis_steps_per_unit[X_AXIS];
	    delta_mm[Y_AXIS] = ((target[X_AXIS]-position[X_AXIS]) - (target[Y_AXIS]-position[Y_AXIS]))/axis_steps_per_unit[Y_AXIS];
	  #endif
	  delta_mm[Z_AXIS] = (target[Z_AXIS]-position_st[Z_AXIS])/config.axis_steps_per_unit[Z_AXIS];
	  delta_mm[E_AXIS] = ((target[E_AXIS]-position_st[E_AXIS])/config.axis_steps_per_unit[E_AXIS])*volumetric_multiplier[active_extruder]*extrudemultiply/100.0;
	  if ( block->steps[X_AXIS] <=dropsegments && block->steps[Y_AXIS] <=dropsegments && block->steps[Z_AXIS] <=dropsegments )
	  {
	    block->millimeters = fabs(delta_mm[E_AXIS]);
	  }
	  else
	  {
	    block->millimeters = sqrt(square(delta_mm[X_AXIS]) + square(delta_mm[Y_AXIS]) + square(delta_mm[Z_AXIS]));
	  }
	  float inverse_millimeters = 1.0/block->millimeters;  // Inverse millimeters to remove multiple divides

	    // Calculate speed in mm/second for each axis. No divide by zero due to previous checks.
	  float inverse_second = feed_rate * inverse_millimeters;

	  int moves_queued=(block_buffer_head-block_buffer_tail + BLOCK_BUFFER_SIZE) & (BLOCK_BUFFER_SIZE - 1);

	  // slow down when de buffer starts to empty, rather than wait at the corner for a buffer refill
	#ifdef SLOWDOWN
	  //  segment time im micro seconds
	  unsigned long segment_time = lround(1000000.0/inverse_second);
	  if ((moves_queued > 1) && (moves_queued < (BLOCK_BUFFER_SIZE * 0.5)))
	  {
	    if (segment_time < config.minsegmenttime)
	    { // buffer is draining, add extra time.  The amount of time added increases if the buffer is still emptied more.
	      inverse_second=1000000.0/(segment_time+lround(2*(config.minsegmenttime-segment_time)/moves_queued));
	      #ifdef XY_FREQUENCY_LIMIT
	         segment_time = lround(1000000.0/inverse_second);
	      #endif
	    }
	  }
	#endif
	  //  END OF SLOW DOWN SECTION


	  // block->nominal_speed = block->millimeters * inverse_second; // (mm/sec) Always > 0
	  block->nominal_speed = feed_rate; // (mm/sec) Always > 0
	  block->nominal_rate = ceil(block->step_event_count * inverse_second); // (step/sec) Always > 0

	  // Calculate and limit speed in mm/sec for each axis
	  float current_speed[4];
	  float speed_factor = 1.0; //factor <=1 do decrease speed
	  for(int i=0; i < 4; i++)
	  {
	    current_speed[i] = delta_mm[i] * inverse_second;
	    if(fabs(current_speed[i]) > config.max_feedrate_mm_s[i])
	      speed_factor = min(speed_factor, config.max_feedrate_mm_s[i] / fabs(current_speed[i]));
	  }

	  // Max segement time in us.
	#ifdef XY_FREQUENCY_LIMIT
	#define MAX_FREQ_TIME (1000000.0/XY_FREQUENCY_LIMIT)
	  // Check and limit the xy direction change frequency
	  unsigned char direction_change = block->direction_bits ^ old_direction_bits;
	  old_direction_bits = block->direction_bits;
	  segment_time = lround((float)segment_time / speed_factor);

	  if((direction_change & (1<<X_AXIS)) == 0)
	  {
	    x_segment_time[0] += segment_time;
	  }
	  else
	  {
	    x_segment_time[2] = x_segment_time[1];
	    x_segment_time[1] = x_segment_time[0];
	    x_segment_time[0] = segment_time;
	  }
	  if((direction_change & (1<<Y_AXIS)) == 0)
	  {
	    y_segment_time[0] += segment_time;
	  }
	  else
	  {
	    y_segment_time[2] = y_segment_time[1];
	    y_segment_time[1] = y_segment_time[0];
	    y_segment_time[0] = segment_time;
	  }
	  long max_x_segment_time = max(x_segment_time[0], max(x_segment_time[1], x_segment_time[2]));
	  long max_y_segment_time = max(y_segment_time[0], max(y_segment_time[1], y_segment_time[2]));
	  long min_xy_segment_time =min(max_x_segment_time, max_y_segment_time);
	  if(min_xy_segment_time < MAX_FREQ_TIME)
	    speed_factor = min(speed_factor, speed_factor * (float)min_xy_segment_time / (float)MAX_FREQ_TIME);
	#endif

	  // Correct the speed
	  // only axial correction. combined x/y moves are handled wrong!
	  //
	  if( speed_factor < 1.0)
	  {
	    for(unsigned char i=0; i < 4; i++)
	    {
	      current_speed[i] *= speed_factor;
	    }
	    block->nominal_speed *= speed_factor;
	    block->nominal_rate *= speed_factor;
	  }

	  // Compute and limit the acceleration rate for the trapezoid generator.
	  float steps_per_mm = block->step_event_count/block->millimeters;
	  if(block->steps[X_AXIS]  == 0 && block->steps[Y_AXIS]  == 0 && block->steps[Z_AXIS]  == 0)
	  {
		// only extruder move
	    block->acceleration_st = ceil(config.retract_acceleration * steps_per_mm); // convert to: acceleration steps/sec^2
	  }
	  else
	  {
	    block->acceleration_st = ceil(config.acceleration * steps_per_mm); // convert to: acceleration steps/sec^2
	    // Limit acceleration per axis
	    if(((float)block->acceleration_st * (float)block->steps[X_AXIS]  / (float)block->step_event_count) > config.axis_steps_per_sqr_second[X_AXIS])
	      block->acceleration_st = config.axis_steps_per_sqr_second[X_AXIS];
	    if(((float)block->acceleration_st * (float)block->steps[Y_AXIS]  / (float)block->step_event_count) > config.axis_steps_per_sqr_second[Y_AXIS])
	      block->acceleration_st = config.axis_steps_per_sqr_second[Y_AXIS];
	    if(((float)block->acceleration_st * (float)block->steps[E_AXIS]  / (float)block->step_event_count) > config.axis_steps_per_sqr_second[E_AXIS])
	      block->acceleration_st = config.axis_steps_per_sqr_second[E_AXIS];
	    if(((float)block->acceleration_st * (float)block->steps[Z_AXIS]  / (float)block->step_event_count ) > config.axis_steps_per_sqr_second[Z_AXIS])
	      block->acceleration_st = config.axis_steps_per_sqr_second[Z_AXIS];
	  }
	  block->acceleration = block->acceleration_st / steps_per_mm;
	  block->acceleration_rate = (long)((float)block->acceleration_st * (16777216.0 / (F_CPU / 8.0)));

		  // Start with a safe speed
	  float vmax_junction = config.max_jerk[X_AXIS]/2;
	  float vmax_junction_factor = 1.0;
	  if(fabs(current_speed[Z_AXIS]) > config.max_jerk[Z_AXIS]/2)
	    vmax_junction = min(vmax_junction, config.max_jerk[Z_AXIS]/2);
	  if(fabs(current_speed[E_AXIS]) > config.max_jerk[E_AXIS]/2)
	    vmax_junction = min(vmax_junction, config.max_jerk[E_AXIS]/2);
	  vmax_junction = min(vmax_junction, block->nominal_speed);
	  float safe_speed = vmax_junction;

	  if ((moves_queued > 1) && (previous_nominal_speed > 0.0001)) {
	    float jerk = sqrt(pow((current_speed[X_AXIS]-previous_speed[X_AXIS]), 2)+pow((current_speed[Y_AXIS]-previous_speed[Y_AXIS]), 2));
	    //    if((fabs(previous_speed[X_AXIS]) > 0.0001) || (fabs(previous_speed[Y_AXIS]) > 0.0001)) {
	    vmax_junction = block->nominal_speed;
	    //    }
	    if (jerk > config.max_jerk[X_AXIS]) {
	      vmax_junction_factor = (config.max_jerk[X_AXIS]/jerk);
	    }
	    // the same for Z axis
	    jerk = fabs(current_speed[Z_AXIS] - previous_speed[Z_AXIS]);
	    if( jerk > config.max_jerk[Z_AXIS]) {
	      vmax_junction_factor= min(vmax_junction_factor, (config.max_jerk[Z_AXIS]/jerk));
	    }
	    // and for extrusion
	    jerk = fabs(current_speed[E_AXIS] - previous_speed[E_AXIS]);
	    if( jerk > config.max_jerk[E_AXIS]) {
	      vmax_junction_factor = min(vmax_junction_factor, (config.max_jerk[E_AXIS]/jerk));
	    }
	    vmax_junction = min(previous_nominal_speed, vmax_junction * vmax_junction_factor); // Limit speed to max previous speed
	  }
	  block->max_entry_speed = vmax_junction;

	#ifdef PRESSURE_ADVANCE
	  // calculate the advance values for this block based on the given advance factor
	  // pressure advance is only applicable during x/y moves, Z-Moves and pure extruder movement is excluded
	  // These values will be optimised within the planner recalculation
	  if (block->steps[X_AXIS] > 0 || block->steps[Y_AXIS] > 0)
	  {
		 // float jerk_e = fabs((e-last_e)* feed_rate * inverse_millimeters );
		  int jerk_count = fabs(feed_rate * config.advance_factor[extruder]);
		  jerk_count = (jerk_count>>3);
		  jerk_count = (jerk_count<<3);
		  if (delta_mm[E_AXIS] > 0)	//active only in extrusion direction, suppress on retract.
		  {
		  block->initial_advance = jerk_count;
		  block->final_advance = -jerk_count;
		  }

	//	    SERIAL_ECHO_START;
	//	   SERIAL_ECHOPGM("advance :");
	//	   SERIAL_ECHO(block->initial_advance);
	//	   SERIAL_ECHOPGM(" advance release :");
	//	   SERIAL_ECHOLN(block->final_advance); wait


	  }
	  last_e = e;
	#endif

	  // Initialize block entry speed. Compute based on deceleration to user-defined MINIMUM_PLANNER_SPEED.
	  double v_allowable = Max_Allowable_Speed(-block->acceleration,MINIMUM_PLANNER_SPEED,block->millimeters);
	  block->entry_speed = min(vmax_junction, v_allowable);

	  // Initialize planner efficiency flags
	  // Set flag if block will always reach maximum junction speed regardless of entry/exit speeds.
	  // If a block can de/ac-celerate from nominal speed to zero within the length of the block, then
	  // the current block and next block junction speeds are guaranteed to always be at their maximum
	  // junction speeds in deceleration and acceleration, respectively. This is due to how the current
	  // block nominal speed limits both the current and next maximum junction speeds. Hence, in both
	  // the reverse and forward planners, the corresponding block junction speed will always be at the
	  // the maximum junction speed and may always be ignored for any speed reduction checks.
	  if (block->nominal_speed <= v_allowable) {
	    block->nominal_length_flag = true;
	  }
	  else {
	    block->nominal_length_flag = false;
	  }
	  block->recalculate_flag = true; // Always calculate trapezoid for new block

	  // Update previous path unit_vector and nominal speed
	  memcpy(previous_speed, current_speed, sizeof(previous_speed)); // previous_speed[] = current_speed[]
	  previous_nominal_speed = block->nominal_speed;



	  Calculate_Trapezoid_for_Block(block, block->entry_speed/block->nominal_speed, safe_speed/block->nominal_speed);

	  // Move buffer head
	  block_buffer_head = next_buffer_head;

	  // Update position
	  memcpy(position_st, target, sizeof(target)); // position_st[] = target[]

	  Recalculate();

	  motor.Wake_Up();
	  my_printer.Reset_MotorTimeout();

}

void Planner::Recalculate()
{
	Reverse_Pass();
	  Forward_Pass();
	  Recalculate_Trapezoids();
}

	// Recalculates the trapezoid speed profiles for all blocks in the plan according to the
	// entry_factor for each junction. Must be called by planner_recalculate() after
	// updating the blocks.
void Planner::Recalculate_Trapezoids()
{
	  int8_t block_index = block_buffer_tail;
	  move_block *current;
	  move_block *next = NULL;

	  while(block_index != block_buffer_head) {
	    current = next;
	    next = &block_buffer[block_index];
	    if (current) {
	      // Recalculate if current block entry or exit junction speed has changed.
	      if (current->recalculate_flag || next->recalculate_flag) {
	        // NOTE: Entry and exit factors always > 0 by all previous logic operations.
	    	  Calculate_Trapezoid_for_Block(current, current->entry_speed/current->nominal_speed,
	        next->entry_speed/current->nominal_speed);
	        current->recalculate_flag = false; // Reset current only to ensure next trapezoid is computed
	      }
	    }
	    block_index = Next_Block_Index( block_index );
	  }
	  // Last/newest block in buffer. Exit speed is set with MINIMUM_PLANNER_SPEED. Always recalculated.
	  if(next != NULL) {
		  Calculate_Trapezoid_for_Block(next, next->entry_speed/next->nominal_speed,
	    MINIMUM_PLANNER_SPEED/next->nominal_speed);
	    next->recalculate_flag = false;
	  }

}
	// planner_recalculate() needs to go over the current plan twice. Once in reverse and once forward. This
	// implements the forward pass.
void Planner::Forward_Pass()
{
	  uint8_t block_index = block_buffer_tail;
	  move_block *block[3] = {
	    NULL, NULL, NULL   };

	  while(block_index != block_buffer_head) {
	    block[0] = block[1];
	    block[1] = block[2];
	    block[2] = &block_buffer[block_index];
	    Forward_Pass_Kernel(block[0],block[1],block[2]);
	    block_index = Next_Block_Index(block_index);
	  }
	  Forward_Pass_Kernel(block[1], block[2], NULL);
}
	// planner_recalculate() needs to go over the current plan twice. Once in reverse and once forward. This
	// implements the reverse pass.
void Planner::Reverse_Pass()
{
		uint8_t block_index = block_buffer_head;

		//Make a local copy of block_buffer_tail, because the interrupt can alter it
		CRITICAL_SECTION_START;
		unsigned char tail = block_buffer_tail;
		CRITICAL_SECTION_END

		if(((block_buffer_head-tail + BLOCK_BUFFER_SIZE) & (BLOCK_BUFFER_SIZE - 1)) > 3)
		{
			block_index = (block_buffer_head - 3) & (BLOCK_BUFFER_SIZE - 1);
			move_block *block[3] =
			{
					NULL, NULL, NULL
			};
			while(block_index != tail)
			{
				block_index = Prev_Block_Index(block_index);
				block[2]= block[1];
				block[1]= block[0];
				block[0] = &block_buffer[block_index];
				Reverse_Pass_Kernel(block[0], block[1], block[2]);
			}
		}
}

	// The kernel called by planner_recalculate() when scanning the plan from first to last entry.

void Planner::Forward_Pass_Kernel(move_block* previous, move_block* current,
		move_block* next)
{
	  if(!previous) {
	    return;
	  }

	  // If the previous block is an acceleration block, but it is not long enough to complete the
	  // full speed change within the block, we need to adjust the entry speed accordingly. Entry
	  // speeds have already been reset, maximized, and reverse planned by reverse planner.
	  // If nominal length is true, max junction speed is guaranteed to be reached. No need to recheck.
	  if (!previous->nominal_length_flag) {
	    if (previous->entry_speed < current->entry_speed) {
	      double entry_speed = min( current->entry_speed,
	    		  Max_Allowable_Speed(-previous->acceleration,previous->entry_speed,previous->millimeters) );

	      // Check for junction speed change
	      if (current->entry_speed != entry_speed) {
	        current->entry_speed = entry_speed;
	        current->recalculate_flag = true;
	      }
	    }
	  }
}
	// The kernel called by planner_recalculate() when scanning the plan from last to first entry.

void Planner::Reverse_Pass_Kernel(move_block* previous, move_block* current,
		move_block* next)
{
	  if(!current) {
	    return;
	  }
	#ifdef PRESSURE_ADVANCE
	      // check if pressure advance of previous block can be compensated with value from current block
	      if (current->initial_advance > 0)
	      {
	    	  if (current->initial_advance > -previous->final_advance)
	    	  {
	    		  current->initial_advance += previous->final_advance;
	    		  previous->final_advance = 0;
	    	  }
	    	  else
	    	  {
	    		  previous->final_advance += current->initial_advance;
	    		  current->initial_advance = 0;
	    	  }
	      }
	#endif

	  if (next) {
	    // If entry speed is already at the maximum entry speed, no need to recheck. Block is cruising.
	    // If not, block in state of acceleration or deceleration. Reset entry speed to maximum and
	    // check for maximum allowable speed reductions to ensure maximum possible planned speed.
	    if (current->entry_speed != current->max_entry_speed) {

	      // If nominal length true, max junction speed is guaranteed to be reached. Only compute
	      // for max allowable speed if block is decelerating and nominal length is false.
	      if ((!current->nominal_length_flag) && (current->max_entry_speed > next->entry_speed)) {
	        current->entry_speed = min( current->max_entry_speed,
	        		Max_Allowable_Speed(-current->acceleration,next->entry_speed,current->millimeters));
	      }
	      else {
	        current->entry_speed = current->max_entry_speed;
	      }
	      current->recalculate_flag = true;

	    }
	  } // Skip last block. Already initialized and set for recalculation.

}
	// Calculates trapezoid parameters so that the entry- and exit-speed is compensated by the provided factors.

void Planner::Calculate_Trapezoid_for_Block(move_block* block,
		float entry_factor, float exit_factor)
{
	  unsigned long initial_rate = ceil(block->nominal_rate*entry_factor); // (step/min)
	  unsigned long final_rate = ceil(block->nominal_rate*exit_factor); // (step/min)

	  // Limit minimal step rate (Otherwise the timer will overflow.)
	  if(initial_rate <120) {
	    initial_rate=120;
	  }
	  if(final_rate < 120) {
	    final_rate=120;
	  }
	  // new distance calculation
	  long acceleration = block->acceleration_st;
	  int32_t accelerate_steps =
	    ceil(Estimate_Acceleration_Distance(initial_rate, block->nominal_rate, acceleration));
	  int32_t decelerate_steps =
	    floor(Estimate_Acceleration_Distance(block->nominal_rate, final_rate, -acceleration));

	  // Calculate the size of Plateau of Nominal Rate.
	  int32_t plateau_steps = block->step_event_count-accelerate_steps-decelerate_steps;

	  // Is the Plateau of Nominal Rate smaller than nothing? That means no cruising, and we will
	  // have to use intersection_distance() to calculate when to abort acceleration and start braking
	  // in order to reach the final_rate exactly at the end of this block.
	  if (plateau_steps < 0) {
	    accelerate_steps = ceil(Intersection_Distance(initial_rate, final_rate, acceleration, block->step_event_count));
	    accelerate_steps = max(accelerate_steps,0); // Check limits due to numerical round-off
	    accelerate_steps = min((uint32_t)accelerate_steps,block->step_event_count);//(We can cast here to unsigned, because the above line ensures that we are above zero)
	    plateau_steps = 0;
	  }


	  // block->accelerate_until = accelerate_steps;
	  // block->decelerate_after = accelerate_steps+plateau_steps;
	  CRITICAL_SECTION_START;  // Fill variables used by the stepper in a critical section
	  if(block->busy == false) { // Don't update variables if block is busy.
	    block->accelerate_until = accelerate_steps;
	    block->decelerate_after = accelerate_steps+plateau_steps;
	    block->initial_rate = initial_rate;
	    block->final_rate = final_rate;

	  }
	  CRITICAL_SECTION_END;

}
// Calculate the steps/s^2 acceleration rates, based on the mm/s^s
//void reset_acceleration_rates()
//{
//	for(int8_t i=0; i < NUM_AXIS; i++)
//        {
//		config.axis_steps_per_sqr_second[i] = config.max_acceleration_units_per_sq_second[i] * config.axis_steps_per_unit[i];
//        }
//}

