/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef planner_h
#define planner_h

#include "MarlinHOH.h"

#ifdef ENABLE_AUTO_BED_LEVELING
#include "vector_3.h"
#endif // ENABLE_AUTO_BED_LEVELING

// This struct is used when buffering the setup for each linear movement "nominal" values are as specified in 
// the source g-code and may never actually be reached if acceleration management is active.
typedef  struct move_block {
  // Fields used by the bresenham algorithm for tracing the line
  long steps[NUM_AXIS];  // Step count along each axis
  unsigned long step_event_count;           // The number of step events required to complete this block
  long accelerate_until;                    // The index of the step event on which to stop acceleration
  long decelerate_after;                    // The index of the step event on which to start decelerating
  long acceleration_rate;                   // The acceleration rate used for acceleration calculation
  unsigned char direction_bits;             // The direction bit set for this block (refers to *_DIRECTION_BIT in config.h)
  unsigned char active_extruder;            // defines the active extruder
  #ifdef PRESSURE_ADVANCE
    volatile long initial_advance;			// the initial pressure advance steps at the beginning of an extrusion
    volatile long final_advance;			// the relaxing prssure advance steps at the end of the block
    volatile int advance_delay;				// the stepping delay between two advance steps
  #endif

  // Fields used by the motion planner to manage acceleration
//  float speed_x, speed_y, speed_z, speed_e;        // Nominal mm/sec for each axis
  float nominal_speed;                               // The nominal speed for this block in mm/sec 
  float entry_speed;                                 // Entry speed at previous-current junction in mm/sec
  float max_entry_speed;                             // Maximum allowable junction entry speed in mm/sec
  float millimeters;                                 // The total travel of this block in mm
  float acceleration;                                // acceleration mm/sec^2
  unsigned char recalculate_flag;                    // Planner flag to recalculate trapezoids on entry junction
  unsigned char nominal_length_flag;                 // Planner flag for nominal speed always reached

  // Settings for the trapezoid generator
  unsigned long nominal_rate;                        // The nominal step rate for this block in step_events/sec 
  unsigned long initial_rate;                        // The jerk-adjusted step rate at start of block  
  unsigned long final_rate;                          // The minimal rate at exit
  unsigned long acceleration_st;                     // acceleration steps/sec^2
  unsigned long fan_speed;
  #ifdef BARICUDA
  unsigned long valve_pressure;
  unsigned long e_to_p_pressure;
  #endif
  bool busy;
} block_t;

#ifdef ENABLE_AUTO_BED_LEVELING
// this holds the required transform to compensate for bed level
extern matrix_3x3 plan_bed_level_matrix;
#endif // #ifdef ENABLE_AUTO_BED_LEVELING

#ifdef PREVENT_DANGEROUS_EXTRUDE
void set_extrude_min_temp(float temp);
#endif

class Planner
{
public:
	Planner();
	virtual ~Planner();

	// Initialize the motion plan subsystem
	void Init();

	FORCE_INLINE bool Blocks_Queued() {
		if (block_buffer_head == block_buffer_tail)
		{
			return false;
		}
		else
			return true;
	};

	FORCE_INLINE uint8_t Moves_Planned()
	{
		  return (block_buffer_head-block_buffer_tail + BLOCK_BUFFER_SIZE) & (BLOCK_BUFFER_SIZE - 1);
	};

	FORCE_INLINE move_block *Get_Current_Block()
	{

		  if (block_buffer_head == block_buffer_tail) {
		    return(NULL);
		  }
		  move_block *block = &block_buffer[block_buffer_tail];
		  block->busy = true;
		  return(block);
	};

	FORCE_INLINE void Discard_Current_Block()
	{
	 	  if (block_buffer_head != block_buffer_tail)
		    block_buffer_tail = (block_buffer_tail + 1) & (BLOCK_BUFFER_SIZE - 1);
	};

	FORCE_INLINE float Estimate_Acceleration_Distance(float initial_rate, float target_rate, float acceleration)
	{
	  if (acceleration!=0) {
	 //   return((target_rate*target_rate-initial_rate*initial_rate)/
	 //     (2.0*acceleration));
		  return (target_rate*target_rate - initial_rate*target_rate)/acceleration;
	  }
	  else {
	    return 0.0;  // acceleration was 0, set acceleration distance to 0
	  }
	};

	// Returns the index of the next block in the ring buffer
	// NOTE: Removed modulo (%) operator, which uses an expensive divide and multiplication.
	FORCE_INLINE int8_t Next_Block_Index(int8_t block_index) {
	  block_index++;
	  if (block_index == BLOCK_BUFFER_SIZE) {
	    block_index = 0;
	  }
	  return(block_index);
	}

	// Returns the index of the previous block in the ring buffer
	FORCE_INLINE int8_t Prev_Block_Index(int8_t block_index) {
	  if (block_index == 0) {
	    block_index = BLOCK_BUFFER_SIZE;
	  }
	  block_index--;
	  return(block_index);
	}


	FORCE_INLINE float Intersection_Distance(float initial_rate, float final_rate, float acceleration, float distance)
	{
	  if (acceleration!=0) {
	    return((2.0*acceleration*distance-initial_rate*initial_rate+final_rate*final_rate)/
	      (4.0*acceleration) );
	  }
	  else {
	    return 0.0;  // acceleration was 0, set intersection distance to 0
	  }
	}

	// Calculates the maximum allowable speed at this point when you must be able to reach target_velocity using the
	// acceleration within the allotted distance.
	FORCE_INLINE float Max_Allowable_Speed(float acceleration, float target_velocity, float distance) {
	  return  sqrt(target_velocity*target_velocity-2*acceleration*distance);
	}


	void Check_Axes_Activity();
	void Set_Position_mm(const float &x, const float &y, const float &z, const float &e);
	void Set_Position_mm(const AxisEnum axis,  const float &value);

	void Buffer_Line(const float &x, const float &y, const float &z, const float &e, float feed_rate, const uint8_t &extruder);

private:
	void Recalculate();
	void Recalculate_Trapezoids();
	void Forward_Pass();
	void Reverse_Pass();

	void Forward_Pass_Kernel(move_block *previous, move_block *current, move_block *next);
	void Reverse_Pass_Kernel(move_block *previous, move_block *current, move_block *next);

	void Calculate_Trapezoid_for_Block(move_block *block, float entry_factor, float exit_factor);

	// The current position of the tool in absolute steps
	long position_st[NUM_AXIS];   //rescaled from extern when axis_steps_per_unit are changed by gcode
	float previous_speed[NUM_AXIS]; // Speed of previous path line segment
	float previous_nominal_speed; // Nominal speed of previous path line segment

	move_block block_buffer[BLOCK_BUFFER_SIZE];            // A ring buffer for motion instfructions
	 unsigned char block_buffer_head;           // Index of the next block to be pushed
	 unsigned char block_buffer_tail;

};

extern class Planner plan;

#endif

